
//Game logic.
//Luke Lenhart, 2008
//See the MPMA Framework license for details on how this code may be used.

#include <iostream>

#include "mpma/base/Timer.h"
#include "mpma/input/Keyboard.h"
#include "mpma/base/MiscStuff.h"
#include "mpma/gfxsetup/Setup.h" //to get size for fps output

#include "Game.h"
#include "Render.h"
#include "NetStuff.h"

//#define DEBUG_STATE_STUFF

namespace
{
    //timer and FPS counter
    MPMA::Timer timer;
    uint fpsCount=0;
    float fpsTime=0;

    float gameTime=0;

    //game elements
    float paddleLeftY=0;
    float paddleRightY=0;
    float paddleLeftYResetMagnitude=0;
    float paddleRightYResetMagnitude=0;

    GEO::Vector2 ball(0, 0);
    GEO::Vector2 ballDir(0, 0);

    const float MIN_BALL_DESIRED_MAG=0.5f;

    enum GameState
    {
        Starting,
        NetSyncBeforePlay,
        Playing,
        NetSyncAfterPlay,
        Ending
    };

    GameState state=Starting;
    float stateTime=0;

    MPMA::Timer controlChangeTimer;

    bool initialControl=true;
    bool inControl=true; //whether we think we are in control of the ball
    bool scoredThisRound=false;

    GameState remoteState=Starting;
    bool remoteThinksHasControl=false;

    GEO::Vector2 lastSaneBallDir=0;
    float badBallTime=0;

    //are we in control of the ball (is it heading towards us?)
    bool AmIInControl()
    {
        if (controlChangeTimer.Step(false)>0.25f) //prevent control jitter
        {
            controlChangeTimer.Step(true);
            {
                if (ball.x()<0)
                    inControl=Game::AmIPLeft;
                else if (ball.x()>0)
                    inControl=Game::AmIPRight;
            }
        }
        
        return inControl;
    }

    //new random ball direction
    void RandomizeBallDir()
    {
        do
        {
            float balldir=rndf()*3.15159f*2;
            ballDir=GEO::Vector2(sin(balldir), cos(balldir));
            ballDir*=6.5f+rndf()*0.5f;
        } while (fabs(ballDir.x())<MIN_BALL_DESIRED_MAG);
    }

    //debug
    void DebugSpewState(const std::string &extra="")
    {
#ifdef DEBUG_STATE_STUFF
        std::cout<<"DEBUG: pl="<<Game::AmIPLeft<<" pr="<<Game::AmIPRight<<" control="<<inControl<<" state="<<state<<" ballxdir="<<ballDir.x()<<" "<<extra<<std::endl;
#endif
    }

    //scene info
    const float paddleX=7;
    const float wallDistance=4.4f;
    const float outOfBoundDistance=8.0f;
}

namespace Game
{
    //exposed state info
    std::string PLeftName="Off the right side";
    std::string PRightName="Off the left side";
    bool AmIPLeft=true;
    bool AmIPRight=true;
    int LeftScore=0;
    int RightScore=0;

    //update the world
    void Update(bool passTime)
    {
        float time=0;
        if (passTime)
        {
            time=(float)timer.Step();
            gameTime+=time;

            //output FPS and score every 10 seconds
            fpsTime+=time;
            ++fpsCount;
            if (fpsTime>=10.0f)
            {
                float actualFPS=fpsCount/fpsTime;
                std::cout<<"FPS (last 10 seconds): "<<actualFPS<<"   Resolution: "<<GFX::GetWindowState()->Width<<"x"<<GFX::GetWindowState()->Height<<"\n";
                std::cout<<PLeftName<<": "<<LeftScore<<"    -    "<<PRightName<<": "<<RightScore<<std::endl;
                fpsTime=0;
                fpsCount=0;

                DebugSpewState();
            }
        }

        //animate light
        Renderer::lightPos.x()=sin(sin(gameTime*0.25f))*10;
        Renderer::lightPos.y()=sin(cos(gameTime*0.333f))*5;
        Renderer::lightPos.z()=-10+sin(gameTime*0.444f)*4;

        //game state
        stateTime+=time;

        //prevent lack of controller
        if (!remoteThinksHasControl && !AmIInControl() && passTime && initialControl)
            inControl=true;

        //game
        if (state==Starting) //setup
        {
            ball=GEO::Vector2(0,0);

            if (stateTime>2)
            {
                DebugSpewState("Done with Starting stage");
                
                scoredThisRound=false;
                ball=0;
                Renderer::ball.radius=Renderer::ballSize;
                stateTime=2;
                state=NetSyncBeforePlay;

                if (AmIInControl() && fabs(ballDir.x())<MIN_BALL_DESIRED_MAG)
                    RandomizeBallDir();
            }
            else
            {
                float mod=stateTime*0.5f;
                Renderer::ball.radius=sin(mod*3.141592f*(5.0f/6))*sin(mod*3.141592f*(5.0f/6))*2*(Renderer::ballSize*2);

                paddleLeftY+=paddleLeftYResetMagnitude*time;
                paddleRightY+=paddleRightYResetMagnitude*time;
            }
        }
        else if (state==NetSyncBeforePlay)
        {
            if (AmIPLeft && AmIPRight) //single player
                state=Playing;
            else if (remoteState>=NetSyncBeforePlay) //multiplayer.. both made it to here
                state=Playing;
        }
        else if (state==Playing) //running
        {
            //prevent stuck ball
            if (ballDir.Length()<MIN_BALL_DESIRED_MAG)
            {
                float chunkyTime=time;
                if (chunkyTime>0.25f) chunkyTime=0.25f;
                badBallTime+=chunkyTime;

                if (badBallTime>1) // we are obviously stuck
                {
                    if (lastSaneBallDir.Length()>=MIN_BALL_DESIRED_MAG)
                        ballDir=lastSaneBallDir;
                    else if (inControl) //intended
                        RandomizeBallDir();
                }
            }
            else
            {
                badBallTime=0;
                lastSaneBallDir=ballDir;
            }

            //controls
            if (time>0.5f) time=0.5f;
            const float speed=5.0f;
            if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::UP) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::W) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::COMMA))
            {
                if (AmIPLeft) paddleLeftY+=speed*time;
                if (AmIPRight) paddleRightY+=speed*time;
            }
            if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::DOWN) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::S) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::O))
            {
                if (AmIPLeft) paddleLeftY-=speed*time;
                if (AmIPRight) paddleRightY-=speed*time;
            }

            if (paddleLeftY>3.95f) paddleLeftY=3.95f;
            if (paddleRightY>3.95f) paddleRightY=3.95f;
            if (paddleLeftY<-3.95f) paddleLeftY=-3.95f;
            if (paddleRightY<-3.95f) paddleRightY=-3.95f;

            //ball update.  rather than doing it correctly, I'm going to cheat and just do stuff in small increments, since it's only 1 object moving here
            const float ballIncAmount=0.01f;
            float ballTimeLeft=time;
            while (ballTimeLeft>0.0f)
            {
                float adv=ballTimeLeft;
                if (adv>ballIncAmount) adv=ballIncAmount;
                ballTimeLeft-=adv;

                //now update it by adv and check collections and stuff and prevent "no x speed" death
                if (AmIInControl() && fabs(ballDir.x())<MIN_BALL_DESIRED_MAG)
                {
                    if (ballDir.x()<0)
                        ballDir.x()-=rndf()*adv*MIN_BALL_DESIRED_MAG*0.5f;
                    else
                        ballDir.x()+=rndf()*adv*MIN_BALL_DESIRED_MAG*0.5f;
                }

                ball+=ballDir*adv;

                //TEMP: control the ball for testing:
                /*if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::HOME))
                    ball.y()+=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::END))
                    ball.y()-=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::DELETE))
                    ball.x()-=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::PAGE_DOWN))
                    ball.x()+=adv;*/

                //borders
                if (ball.y()>wallDistance)
                {
                    ball.y()=wallDistance-0.0001f;
                    ballDir.y()*=-1;
                }

                if (ball.y()<-wallDistance)
                {
                    ball.y()=-wallDistance+0.0001f;
                    ballDir.y()*=-1;
                }

                //paddle (hacky work, not technically accurate)
                //for this we use a point in the direction of the ball from the paddle that roughly follows the shape of the paddle to detect
                if (AmIPLeft)
                { //left
                    GEO::Vector2 paddleLeftPos(-paddleX, paddleLeftY);
                    GEO::Vector2 reachToBall=(ball-paddleLeftPos).Normal() * Renderer::paddleLeft.radius * 1.3f;
                    reachToBall.x()/=Renderer::paddleLeft.ymultiplier;
                    reachToBall.y()*=2;
                    GEO::Vector2 reachPoint=paddleLeftPos+reachToBall;
                    if (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius+0.2f)
                    {
                        GEO::Vector2 effectiveNormal=reachToBall;
                        effectiveNormal.x()*=10;

                        if (GEO::VecDot(ballDir.Normal(), (ball-paddleLeftPos).Normal())<0) //different directions, do reflect
                            ballDir=GEO::VecReflect(ballDir, effectiveNormal.Normal());
                        else //same direction.. push it
                            ballDir+=(ball-paddleLeftPos)*adv;

                        reachPoint=paddleLeftPos+reachToBall*1.1f;
                        while (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius)
                            ball+=ballDir*0.0001f;
                    }
                }
                if (AmIPRight)
                { //right
                    GEO::Vector2 paddleRightPos(paddleX, paddleRightY);
                    GEO::Vector2 reachToBall=(ball-paddleRightPos).Normal() * Renderer::paddleRight.radius * 1.3f;
                    reachToBall.y()*=2;
                    reachToBall.x()/=Renderer::paddleRight.ymultiplier;
                    GEO::Vector2 reachPoint=paddleRightPos+reachToBall;
                    if (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius+0.2f)
                    {
                        GEO::Vector2 effectiveNormal=reachToBall;
                        effectiveNormal.x()*=10;

                        if (GEO::VecDot(ballDir.Normal(), (ball-paddleRightPos).Normal())<0) //different directions, do reflect
                            ballDir=GEO::VecReflect(ballDir, effectiveNormal.Normal());
                        else //same direction.. push it
                            ballDir+=(ball-paddleRightPos)*adv;

                        reachPoint=paddleRightPos+reachToBall*1.1f;
                        while (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius)
                            ball+=ballDir*0.0001f;
                    }
                }

                //out of bounds
                if (ball.x()<-outOfBoundDistance || ball.x()>outOfBoundDistance)
                {
                    if (AmIInControl() || remoteState==NetSyncAfterPlay)
                    {
                        if (!scoredThisRound)
                        {
                            if (ball.x()<0)
                                ++RightScore;
                            else
                                ++LeftScore;
                        }
                        std::cout<<"Score!\n"<<PLeftName<<": "<<LeftScore<<"    -    "<<PRightName<<": "<<RightScore<<std::endl;
                        scoredThisRound=true;
                        
                        stateTime=0;
                        state=NetSyncAfterPlay;

                        //force a network update now
                        PongNet::Update(true, false);
                        ballDir=0;
                        
                        DebugSpewState("Done with Playing stage");
                        break;
                    }
                }
            }
        }
        else if (state==NetSyncAfterPlay)
        {
            if (AmIPLeft && AmIPRight) //single player
                state=Ending;
            else if (remoteState>=NetSyncAfterPlay || remoteState<=NetSyncBeforePlay) //multiplayer.. both made it to here
                state=Ending;
        }
        else if (state==Ending) //ball went off the edge
        {
            if (stateTime>1.0f)
            {
                DebugSpewState("Done with Ending stage");

                stateTime=0;
                state=Starting;
                paddleLeftYResetMagnitude=-paddleLeftY/2;
                paddleRightYResetMagnitude=-paddleRightY/2;
            }
            else
            {
                Renderer::ball.radius=(1-stateTime)*Renderer::ballSize;
            }
        }

        //sync renderer to state
        Renderer::paddleLeft.pos.y()=paddleLeftY;
        Renderer::paddleLeft.pos.x()=-paddleX;
        Renderer::paddleRight.pos.y()=paddleRightY;
        Renderer::paddleRight.pos.x()=paddleX;
        Renderer::ball.pos.x()=ball.x();
        Renderer::ball.pos.y()=ball.y();
    }

    // -- interface for networking stuff to poke us with

    //give up control of the ball (if its state is not determined)
    void YieldInitialControl()
    {
        inControl=false;
        initialControl=false;
    }

    //retrieves how we see the world
    NetGameState GetNetState()
    {
        NetGameState net;

        net.inControl=AmIInControl();
        net.gameStage=state;
        net.ballPos=ball;
        net.ballDir=ballDir;
        net.paddleLeftY=paddleLeftY;
        net.paddleRightY=paddleRightY;

        return net;
    }

    //updates how we se ethe world - we will ignore part of this
    void UpdateNetState(const NetGameState &net)
    {
        //if they are in control, listen to their ball position reports
        if (!AmIInControl())
        {
            if (ballDir.Length()<MIN_BALL_DESIRED_MAG) //complete sync
            {
                ball=net.ballPos;
                ballDir=net.ballDir;
            }
            else //adjust
            {
                ball=(ball+net.ballPos)/2;
                //ballDir=(ballDir+net.ballDir)/2; //this can cause some freak ball "slow down" near wall collisions.
            }
        }

        //they can only fill in their own paddle pos
        if (!AmIPLeft)
            paddleLeftY=net.paddleLeftY;

        if (!AmIPRight)
            paddleRightY=net.paddleRightY;

        remoteState=(GameState)net.gameStage;
        remoteThinksHasControl=net.inControl!=0;

        Update(false);
    }
}
