//Networking code for raypong.
//Luke Lenhart, 2008
//See the MPMA Framework license for details on how this code may be used.

#include "mpma/base/Vary.h"
#include "mpma/base/Thread.h"
#include "mpma/base/Timer.h"
#include "mpma/net/TCP.h"
#include "mpma/net/UPNP.h"

#include <iostream>
#include <memory.h>

#include "Game.h"

namespace
{
    const int GAME_PORT=0x11AB; //BALL!  This is also decimal value 4523
    const float MAX_NET_FPS=30;

    //our connection to the other side
    NET::TCPClient *tcp=0;

    //timing information
    MPMA::Timer sendTimer;

    // -- communication structures and constants
    const uint16 PROTOCOL_VERSION=1;
    const uint8 MAX_NAME_LENGTH=30;

    //helpers
    bool Receive1Byte(uint8 &b)
    {
        std::vector<uint8> bytes;
        while (tcp->Receive(bytes, 1)==0)
        {
            if (!tcp->IsConnected())
                return false;
            MPMA::Sleep(1);
        }
        b=*(uint8*)&bytes[0];
        return true;
    }

    bool Receive2Bytes(uint16 &w)
    {
        std::vector<uint8> bytes;
        while (tcp->Receive(bytes, 2)==0)
        {
            if (!tcp->IsConnected())
                return false;
            MPMA::Sleep(1);
        }
        w=*(uint16*)&bytes[0];
        return true;
    }

    bool ReceiveString(std::string &s, uint length)
    {
        std::vector<uint8> bytes;
        while (tcp->Receive(bytes, length)==0)
        {
            if (!tcp->IsConnected())
                return false;
            MPMA::Sleep(1);
        }
        s.resize(length);
        memcpy(&s[0], &bytes[0], length);
        return true;
    }
};

// --

namespace PongNet
{
    //info
    bool IsNetGame=false;

    //cleans up any net resources
    void Cleanup()
    {
        if (tcp)
            NET::TCPClient::FreeClient(tcp);
        tcp=0;
    }

    //Negotiates a connection with another player.
    bool NegotiateConnection()
    {
        //ask them their name
        std::string myName;
        while (myName.empty())
        {
            std::cout<<"Enter your name: ";
            std::cin>>myName;
        }

        if (myName.size()>MAX_NAME_LENGTH)
        {
            myName.resize(MAX_NAME_LENGTH);
            std::cout<<"Name truncated to: "+myName<<std::endl;
        }
        
        //ask them whether we are the host or the client for the initial connection
        int option;
        do
        {
            std::cout<<"\nNetwork Setup:\n";
            std::cout<<" 1. Listen for connection.\n";
            std::cout<<" 2. Connect to remote player."<<std::endl;
            std::string input;
            std::cin>>input;
            option=MPMA::Vary(input);
        } while (!(option==1 || option==2));
        bool isHost=option==1;

        //establish a connection now. the result of this is we will have a TCPClient that lets us talk to the other side.
        if (isHost) //listen for one
        {
            NET::TCPServer *server=NET::TCPServer::Listen(GAME_PORT);
            if (!server)
            {
                std::cout<<"Not able to listen on game port: "<<GAME_PORT<<std::endl;
                return false;
            }

            if (NET::UPNP::IsAvailable())
                std::cout<<"Internet IP: "<<NET::UPNP::GetInternetFacingIP()<<"\n";
            else
                std::cout<<"UPNP is not available, so unable to determine internet IP."<<std::endl;
            std::cout<<"Local IP: "<<NET::GetLocalIP()<<"\n";
            std::cout<<"Listening for connection on port "<<GAME_PORT<<std::endl;

            while (!tcp)
            {
                tcp=server->GetNextConnection();
                MPMA::Sleep(100);
            }

            NET::TCPServer::FreeServer(server);
        }
        else //establish one
        {
            std::string remoteHost;
            while (remoteHost.empty())
            {
                std::cout<<"Enter remote host IP: ";
                std::cin>>remoteHost;
            }

            NET::Address remoteAddr=NET::Address(remoteHost);
            remoteAddr.SetPort(GAME_PORT);

            while (!tcp)
            {
                std::cout<<"Attempting to connect to remote: "<<remoteAddr.GetIPPortString()<<"..."<<std::endl;
                tcp=NET::TCPClient::Connect(remoteAddr);
                if (!tcp)
                {
                    std::cout<<"Connection attempt failed."<<std::endl;
                    MPMA::Sleep(1000);
                }
            }
        }

        //exchange version info
        std::cout<<"Exchanging protocol...\n";
        tcp->Send(&PROTOCOL_VERSION, 2);

        uint16 theirVersion=0;
        if (!Receive2Bytes(theirVersion))
        {
            std::cout<<"Lost connection exchanging protocol version.\n";
            return false;
        }        
        if (theirVersion!=PROTOCOL_VERSION)
        {
            std::cout<<"Not able to play with remote side because they are using protocol version "<<theirVersion<<" but we are using protocol version "<<PROTOCOL_VERSION<<std::endl;
            Cleanup(); //terminate their connection now rather than later
            return false;
        }        

        //exchange names
        std::cout<<"Exchanging name...\n";
        uint8 nameLen=myName.size();
        tcp->Send(&nameLen, sizeof(nameLen));
        tcp->Send(&myName[0], myName.length());     

        if (!Receive1Byte(nameLen))
            return false;

        if (nameLen>MAX_NAME_LENGTH || nameLen==0)
        {
            std::cout<<"Other side sent a bad name length.\n";
            return false;
        }

        std::string theirName;
        if (!ReceiveString(theirName, nameLen))
            return false;

        //now set up the game state based on what we know.
        if (isHost)
        {
            Game::AmIPLeft=true;
            Game::AmIPRight=false;
            Game::PLeftName=myName;
            Game::PRightName=theirName;
        }
        else
        {
            Game::AmIPLeft=false;
            Game::AmIPRight=true;
            Game::PLeftName=theirName;
            Game::PRightName=myName;
            Game::YieldInitialControl();
        }

        IsNetGame=true;

        //
        tcp->EnableSendCoalescing(false);
        std::cout<<"Done exchanging initial information."<<std::endl;
        return true;
    }

    //Updates network stuff and possibly changes the game state based on it.  If this returns false we lost connection.
    bool Update(bool force, bool allowRecieve)
    {
        if (!tcp || !tcp->IsConnected())
            return false;

        //send our state if needed
        if (force || sendTimer.Step(false)>=(1.0/MAX_NET_FPS))
        {
            sendTimer.Step(true);

            Game::NetGameState state=Game::GetNetState();
            tcp->Send(&state, Game::NetGameStateSize);
        }

        //receive any state info they sent
        if (allowRecieve)
        {
            std::vector<uint8> bytes;
            while (tcp->Receive(bytes, Game::NetGameStateSize))
            {
                Game::NetGameState state;
                memcpy(&state, &bytes[0], Game::NetGameStateSize);
                Game::UpdateNetState(state);

                bytes.clear();
            }
        }

        return tcp->IsConnected();
    }
}
