//! \file Profiler.h \brief Profile the speed of sections of code.
//!
//!The names of a profile are "not" case sensitive.  The results of the profile will be written to the file _profile.txt after the progam ends.  Unless //!PROFILE_RELEASE is defined, profiles will only be taken for a debug build and release builds the macros resolve to nothing.

/*
An example of profiling a section of code:
  ProfileStart("Code Bit 1");
  DoSomething();
  DoSomethingElse();
  ProfileStop("Code Bit 1");

An example of automatically profiling the scope of a function:
  void SomeFunction()
  {
      ProfileScope("name of a profile");
      //..normal code and stuff....
  }
*/
/*
Written by Luke Lenhart (2002)
 Updated in late 2004 to be thread-safe, and to use stl
 Updated 2005 for auto-scope profiling
 Updated 2007 for multiple platforms
See /docs/License.txt for details on how this code may be used.
*/

#ifndef _CProfiler_
#define _CProfiler_

#include "Setup.h"

//macro replacement
#if defined(_DEBUG) || defined(PROFILE_RELEASE)
    //!Begin profiling a section of code.
    #define ProfileStart(name) _instProfile._ProfileStart(name,__FILE__)
    //!End profiling a section of code.
    #define ProfileStop(name) _instProfile._ProfileStop(name,__FILE__)
    //!Profile a scope of code.
    #define ProfileScope(name) CAutoProfileHelper _auto_scope_profiler(name); ProfileStart(name)
#else
    #define ProfileStart(name)
    #define ProfileStop(name)
    #define ProfileScope(name)
#endif


//
#if defined(_DEBUG) || defined(PROFILE_RELEASE)

#include <string>
#include <list>
#include "Locks.h"
#include "Types.h"
#include "Timer.h"

//main profiler
class CProfiler
{
public:
    CProfiler();
    virtual ~CProfiler();

    void _ProfileStart(const std::string &pName, const std::string &fName);
    void _ProfileStop(const std::string &pName, const std::string &fName);

private:
    struct SProfile
    {
        std::string name; //name for this profile
        std::string file; //filename for this profile

        double timeTotal; //total time spent
        double timeMax; //largest time spent
        double timeMin; //smallest time spent

        MPMA::Timer timeStart; //time of last profile start

        int samples; //number of samples taken
        
        bool started;

        inline bool operator <(const SProfile &o) { return timeTotal>o.timeTotal; }
    };

    std::list<SProfile> llProf; //profile list

    MPMA::Timer timeStart; //time class was created

    //

    void Write(FILE* f, double val);
    void Write(FILE* f, const char* str);
    void Write(FILE* f, int num);

    SProfile* FindProfile(const std::string &name);

    void WriteProfilesToFile(FILE *f, double finalTime);
    void DeallocProfiles();

    void Error(const char *lpzMsg);

    //thread safety
    MPMA::MutexLock crit;
};

#endif

//access to profiler object
#if defined(_DEBUG) || defined(PROFILE_RELEASE)
 extern CProfiler _instProfile;
#endif

 //auto-scope profiler (exception-safe and return-safe) -- use macro, not this!!
#if defined(_DEBUG) || defined(PROFILE_RELEASE)
class CAutoProfileHelper
{
public:
    inline CAutoProfileHelper(const std::string &profileName) { name=profileName; }
    inline ~CAutoProfileHelper() { ProfileStop(name); }

private:
    std::string name;
};
#endif

#endif //#ifdef _CProfiler_
