//!\file Setup.h Window system and OpenGL setup
//Luke Lenhart, 2008
//See /docs/License.txt for details on how this code may be used.

#pragma once
#include "../base/Types.h"
#include <string>

//!Graphics
namespace GFX
{
    //!Represents a the size of a window or screen.
    struct ScreenSize
    {
        //!Horizontal pixels contained within the window.
        int Width;

        //!Vertical pixels contained within the window.
        int Height;
    };

    //!Represents the state of graphics system and its related window.  Everything in here can be both changed or read from.
    struct GraphicsSetup: public ScreenSize
    {
        //!The name of the window.
        std::string Name;

        //!Whether the window takes over the wholescreen.  Width and Height will be automatically adjusted accordingly.
        bool FullScreen;

        //!Whether the window can be resized by the user (this is ignored if FullScreen is true).
        bool Resizable;

        //!Whether the window is currently minimized.
        bool Minimized;

        //TODO: stencil buffer and vsync options

        inline GraphicsSetup()
        {
            Name="App";
            FullScreen=false;
            Resizable=true;
            Minimized=false;
            Width=800;
            Height=600;
        }
    };

    //!Creates or updates the state of the applications window and sets up OpenGL.  Returns true if changing to the specified settings was successful, false if it failed.  This must be called at least once before OpenGL can be used.  Only the thread that called this to create the window may make OpenGL calls.
    bool SetupWindow(const GraphicsSetup &newState);

    //!Closes the window and shuts down the graphics system.  This must be called from the same thread that called SetupWindow to create the initial window.
    void ShutdownWindow();

    //!Retrieves the current state of the window, or 0 if not running.
    const GraphicsSetup* GetWindowState();

    //!Handles any window system events, and exposes the backbuffer.  You should normally call this every frame.  You should check GetWindowState after calling this to verify that the window still exists.  This must be called from the same thread that called SetupWindow to create the initial window.
    void UpdateWindow();

    //!Returns whether the window currently has the user's focus.
    bool WindowHasFocus();

    //!Returns the resolution of the current screen.
    ScreenSize GetCurrentScreenSize();

    //TODO: An api to get a list of supported screen resolutions
    //TODO: An api to change the current screen resolution
}
