//!\file Setup.cpp Window system and OpenGL setup
//Luke Lenhart, 2008
//See /docs/License.txt for details on how this code may be used.

#include "../Setup.h"
#include "../../base/DebugRouter.h"
#include <X11/X.h>
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glu.h>
#include <list>
#include <vector>

//internal event hookups to other parts that need window and update events
namespace GFX_INTERNAL
{
    std::list<void (*)()> *pUpdateCallbacks=0;
    std::list<void (*)(XEvent*)> *pEventCallbacks=0;

    //internal use only - used to hook up window update callbacks needed by other components
    void AddUpdateCallback(void (*callback)())
    {
        //we do not free this, it is a one time ever alloc
        if (pUpdateCallbacks==0)
            pUpdateCallbacks=new std::list<void (*)()>();

        pUpdateCallbacks->push_back(callback);
    }

    //internal use only - used to hook up window event callbacks needed by other components
    void AddEventCallback(void (*callback)(XEvent*))
    {
        //we do not free this, it is a one time ever alloc
        if (pEventCallbacks==0)
            pEventCallbacks=new std::list<void (*)(XEvent*)>();

        pEventCallbacks->push_back(callback);
    }
}
using namespace GFX_INTERNAL;

namespace
{
    bool isInited=false;
    bool isShutdownHooked=false;
    GFX::GraphicsSetup state;

    Display *display=0;
    XVisualInfo *visualInfo=0;
    Window window=0;
    bool ignoreSizeEvent=false;

    Atom atomDeleteWindow=None;
    Atom atomChangeState=None;

    GLXContext context=0;

    bool windowHasFocus=false;

    //Make the initial window
    bool CreateWindow()
    {
        //set up X
        display=XOpenDisplay(0);
        if (display==0)
        {
            MPMA::ErrorReport()<<"XOpenDisplay failed.\n";
            return false;
        }

        Window rootWindow=XDefaultRootWindow(display);

        //pick the attributes we want
        GLint visualAttributes[]=
        {
            GLX_RGBA,
            GLX_DEPTH_SIZE, 24,
            GLX_DOUBLEBUFFER,
            None
        };
        visualInfo=glXChooseVisual(display, 0, visualAttributes);
        if (visualInfo==0)
        {
            MPMA::ErrorReport()<<"glXChooseVisual failed.\n";
            return false;
        }

        //set up the window
        Colormap colorMap=XCreateColormap(display, rootWindow, visualInfo->visual, AllocNone);

        XSetWindowAttributes swa;
        swa.colormap=colorMap;
        swa.event_mask=StructureNotifyMask|FocusChangeMask|KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask;

        window=XCreateWindow(display, rootWindow, 0, 0, state.Width, state.Height, 0, visualInfo->depth, InputOutput, visualInfo->visual, CWColormap|CWEventMask, &swa);

        if (window)
            XMapWindow(display, window);
        else
        {
            MPMA::ErrorReport()<<"XCreateWindow failed.\n";
            return false;
        }

        //set us up to recieve a notifications for certain window manager things
        std::vector<Atom> atoms;

        atomDeleteWindow=XInternAtom(display, "WM_DELETE_WINDOW", false);
        if (atomDeleteWindow!=None)
            atoms.push_back(atomDeleteWindow);
        else
            MPMA::ErrorReport()<<"Unable to register WM_DELETE_WINDOW.  App may not close cleanly if the user closes the window.\n";

        atomChangeState=XInternAtom(display, "WM_CHANGE_STATE", false);
        if (atomChangeState!=None)
            atoms.push_back(atomChangeState);
        else
            MPMA::ErrorReport()<<"Unable to register WM_CHANGE_STATE.  App may not react to minimize/restore correctly.\n";

        if (!atoms.empty())
            XSetWMProtocols(display, window, &atoms[0], atoms.size());

        //set up OpenGL
        context=glXCreateContext(display, visualInfo, 0, GL_TRUE);
        if (context)
            glXMakeCurrent(display, window, context);
        else
        {
            MPMA::ErrorReport()<<"glXCreateContext failed.\n";
            return false;
        }

        windowHasFocus=true;

        return true;
    }

    //callback for framework is shutting down
    void GFXFrameworkShutdownCallback()
    {
        GFX::ShutdownWindow();
    }
}

namespace GFX
{
    //Creates or updates the state of the applications window and sets up OpenGL.  Returns true if changing to the specified settings was successful, false if it failed.  This must be called at least once before OpenGL can be used.
    //TODO: better error checking on state updates
    bool SetupWindow(const GraphicsSetup &newState)
    {
        //MPMA::ErrorReport()<<"setup\n";

        //sanitize input
        //TODO:

        //initial setup
        bool firstTime=false;
        if (!isInited)
        {
            firstTime=true;

            //hook our setup
            if (!isShutdownHooked)
            {
                isShutdownHooked=true;
                MPMA::Internal_AddShutdownCallback(GFXFrameworkShutdownCallback, 8000);
            }

            //make the initial window
            state=newState;
            if (!CreateWindow())
            {
                ShutdownWindow();
                return false;
            }

            isInited=true;
        }

        //adjust name
        if (firstTime || state.Name!=newState.Name)
        {
            XStoreName(display, window, newState.Name.c_str());
        }

        //handle display FullScreen change
        if (state.FullScreen!=newState.FullScreen)
        {
            //TODO:
            MPMA::ErrorReport()<<"TODO: Handle FullScreen change.\n";
        }

        //adjust size
        if (state.Width!=newState.Width || state.Height!=newState.Height)
        {
            XWindowChanges change;
            change.width=newState.Width;
            change.height=newState.Height;
            XConfigureWindow(display, window, CWWidth|CWHeight, &change);

            ignoreSizeEvent=true;
        }

        //adjust resizability
        if (firstTime || state.Resizable!=newState.Resizable || state.FullScreen!=newState.FullScreen)
        {
            XSizeHints *xsh=XAllocSizeHints();
            if (!newState.Resizable || newState.FullScreen) //fixed size
            {
                xsh->min_width=newState.Width;
                xsh->min_height=newState.Height;
                xsh->max_width=newState.Width;
                xsh->max_height=newState.Height;
            }
            else //user controls size
            {
                xsh->min_width=1;
                xsh->min_height=1;
                xsh->max_width=8000;
                xsh->max_height=8000;
            }
            xsh->flags=PMinSize|PMaxSize;
            XSetWMNormalHints(display, window, xsh);
            XFree(xsh);
        }

        //adjust iconic state
        if (firstTime || state.Minimized!=newState.Minimized)
        {
            if (newState.Minimized)
                XIconifyWindow(display, window, DefaultScreen(display));
            else
                XRaiseWindow(display, window);
        }

        //flush the x event queue and update once
        state=newState;
        UpdateWindow();

        ignoreSizeEvent=false;
        return true;
    }

    //Closes the window and shuts down the graphics system.
    void ShutdownWindow()
    {
        //clean up OpenGL
        if (context!=0)
        {
            glXMakeCurrent(display, None, 0);
            glXDestroyContext(display, context);
            context=0;
        }

        //clean up X
        if (window!=0)
        {
            XDestroyWindow(display, window);
            window=0;
        }

        if (display!=0)
        {
            XFlush(display);
            XCloseDisplay(display);
            display=0;
        }

        atomDeleteWindow=None;
        atomChangeState=None;
        isInited=false;
        windowHasFocus=false;
    }

    //Retrieves the current state of the window. (or null pointer if not running)
    const GraphicsSetup* GetWindowState()
    {
        if (isInited)
            return &state;
        else
            return 0;
    }

    //Returns the resolution of the current screen.
    ScreenSize GetCurrentScreenSize()
    {
        ScreenSize size;
        size.Width=1;
        size.Height=1;

        if (isInited)
        {
            size.Width=XDisplayWidth(display, DefaultScreen(display));
            size.Height=XDisplayHeight(display, DefaultScreen(display));
        }

        return size;
    }

    //Handles any window system events.  You should normally call this every frame.
    void UpdateWindow()
    {
        if (!isInited)
            return;

        //callback to everyone that we are about to update
        if (pUpdateCallbacks)
        {
            for (std::list<void (*)()>::iterator i=pUpdateCallbacks->begin(); i!=pUpdateCallbacks->end(); ++i)
                (*i)();
        }

        //see if anything happened
        while (XPending(display))
        {
            XEvent event;
            XNextEvent(display, &event);

            if (event.type==ClientMessage) //window manager stuff
            {
                if (event.xclient.format==32)
                {
                    //if (atomDeleteWindow!=None && event.xclient.message_type==atomDeleteWindow) //user asked us to close -- why does this line not work?
                    if (atomDeleteWindow!=None && (Atom)event.xclient.data.l[0]==atomDeleteWindow) //user asked us to close
                    {
                        ShutdownWindow();
                        break; //we're shutdown, don't try to process anything more
                    }
                    else if (atomChangeState!=None && event.xclient.message_type==atomChangeState) //some sorta state change
                    //else if (atomChangeState!=None && (Atom)event.xclient.data.l[0]==atomChangeState) //some sorta state change
                    {
                        if (event.xclient.data.l[0]==IconicState)
                            state.Minimized=true;
                        else //if... what?
                            state.Minimized=false;
                    }
                }
            }
            else if (event.type==ConfigureNotify) //resize
            {
                if (!ignoreSizeEvent)
                {
                    state.Width=event.xconfigure.width;
                    state.Height=event.xconfigure.height;

                    if (state.Width<1) state.Width=1;
                    if (state.Height<1) state.Height=1;
                }
            }
            else if (event.type==FocusIn) //gain focus
                windowHasFocus=true;
            else if (event.type==FocusOut) //lost focus
                windowHasFocus=false;

            //callback to everyone that wants to see events
            if (pEventCallbacks)
            {
                for (std::list<void (*)(XEvent*)>::iterator i=pEventCallbacks->begin(); i!=pEventCallbacks->end(); ++i)
                    (*i)(&event);
            }
        }

        //update backbuffer
        if (isInited)
        {
            glXSwapBuffers(display, window);
        }
    }

    //Returns whether the window currently has the user's focus.
    bool WindowHasFocus()
    {
        if (!isInited)
            return false;

        return windowHasFocus && !state.Minimized;
    }
}
