//!\file Unified.cpp Unified input platform-common code.
//Luke Lenhart, 2008
//See /docs/License.txt for details on how this code may be used.

#include "Unified.h"
#include "Keyboard.h"
#include "Mouse.h"
#include "../base/DebugRouter.h"

namespace GFX_INTERNAL
{
    void AddUpdateCallback(void (*callback)());
}
using namespace GFX_INTERNAL;

//locals
namespace
{
    const std::string badDevice="??? Unknown Device Type ???";
    const std::string tempMessage="TODO: this type";

    bool isCurrentlyPressedValid=false;
    std::vector<INPUT::UnifiedButton> currentlyPressed;
    bool isNewlyPressedValid=false;
    std::vector<INPUT::UnifiedButton> newlyPressed;

    //called when a frame update starts, wipe our button states
    void WindowUpdateCallback()
    {
        isCurrentlyPressedValid=false;
        isNewlyPressedValid=false;
    }

    //hookup callbacks
    class AutoInitInput
    {
    public:
        AutoInitInput()
        {
            AddUpdateCallback(WindowUpdateCallback);
        }
    } autoInitInput;
}

namespace INPUT
{
    //Returns whether the specific button is currently pressed down
    bool IsButtonDown(const UnifiedButton &button)
    {
        if (button.Type==KEYBOARD_DEVICE)
            return KEYBOARD::IsKeyDown(button.Data);
        else if (button.Type==MOUSE_DEVICE)
            return MOUSE::IsButtonDown(button.Data);
        else if (button.Type==GAMEPAD_DEVICE)
            return false; //TODO
        else
        {
            MPMA::ErrorReport()<<"Unknown device type passed to Unified IsButtonDown.\n";
            return false;
        }
    }

    //Returns a list of buttons that are currently pressed down
    const std::vector<UnifiedButton>& GetCurrentlyPressedButtons()
    {
        if (isCurrentlyPressedValid)
            return currentlyPressed;

        //rebuild the current list from all devices
        isCurrentlyPressedValid=true;
        currentlyPressed.clear();

        const std::vector<uint8> &kbKeys=INPUT::KEYBOARD::GetCurrentlyPressedKeys();
        for (std::vector<uint8>::const_iterator i=kbKeys.begin(); i!=kbKeys.end(); ++i)
            currentlyPressed.push_back(UnifiedButton(KEYBOARD_DEVICE, *i));

        const std::vector<uint8> &mouseButtons=INPUT::MOUSE::GetCurrentlyPressedButtons();
        for (std::vector<uint8>::const_iterator i=mouseButtons.begin(); i!=mouseButtons.end(); ++i)
            currentlyPressed.push_back(UnifiedButton(MOUSE_DEVICE, *i));

        //TODO: gamepad

        return currentlyPressed;
    }

    //Returns a list of buttons that were pressed since the last frame
    const std::vector<UnifiedButton>& GetNewlyPressedButtons()
    {
        if (isNewlyPressedValid)
            return newlyPressed;

        //rebuild the current list from all devices
        isNewlyPressedValid=true;
        newlyPressed.clear();

        const std::vector<uint8> &kbKeys=INPUT::KEYBOARD::GetNewlyPressedKeys();
        for (std::vector<uint8>::const_iterator i=kbKeys.begin(); i!=kbKeys.end(); ++i)
            newlyPressed.push_back(UnifiedButton(KEYBOARD_DEVICE, *i));

        const std::vector<uint8> &mouseButtons=INPUT::MOUSE::GetNewlyPressedButtons();
        for (std::vector<uint8>::const_iterator i=mouseButtons.begin(); i!=mouseButtons.end(); ++i)
            newlyPressed.push_back(UnifiedButton(MOUSE_DEVICE, *i));

        //TODO: gamepad

        return newlyPressed;
    }

    //Retrieves a user-friendly name for a button
    const std::string& GetFriendlyName(const UnifiedButton &button)
    {
        if (button.Type==KEYBOARD_DEVICE)
            return KEYBOARD::GetFriendlyName((uint8)button.Data);
        else if (button.Type==MOUSE_DEVICE)
            return MOUSE::GetFriendlyName((uint8)button.Data);
        else if (button.Type==GAMEPAD_DEVICE)
            return tempMessage; //TODO
        else
        {
            MPMA::ErrorReport()<<"Unknown device type passed to Unified GetFriendlyName.\n";
            return badDevice;
        }
    }
}
